/**
 @file BMSFile.m
 @brief BMS 노트/파일 인터페이스 구현
 
 Created by Hoppang on 08. 12. 20.
 */

#import "BMSFile.h"
#import "EDController.h"

/*
 BMSNote는 데이터만 있으므로 synthesize만 해 주면 되겠다.
 */
@implementation BMSNote

@synthesize indexOfInstrument;
@synthesize appearanceTime;
@synthesize length;
@synthesize type;
@synthesize channel;
@synthesize track;
@synthesize audio;

@end

@implementation BMSFile

@synthesize bmspath;
@synthesize stageFile;
@synthesize length;

/**
 @brief BMS 파일을 읽고 파싱해서 데이터를 저장한다.
 @param path BMS파일의 경로
 
 수행순서:
 -# 디스크에 존재하는 BMS 파일을 연다
 -# BMS 파일명의 절대경로에서 파일이름만 잘라서 두 부분을 따로 저장한다
 -# 전체 BMS 텍스트를 줄 단위로 자른다
 -# 파싱 메서드 호출
 -# 필요한 값을 설정하고 상위 init 메서드를 호출
 -# 반환
 */
- (id)initWithFile:(NSString *)path edController:(EDController *)aEDController
{
	int i;
	for(i = 0; i < 255; i++)
		audioFiles[i] = nil;
	
	// BMS 파일의 문자열 전체를 읽어서 메모리에 저장.
	NSError *error;
	NSString *rawBmsText = [NSString stringWithContentsOfFile:path
													 encoding:NSASCIIStringEncoding
														error:&error];
	
	// 절대경로로 전달된 BMS파일명에서 마지막(파일이름)을 떼낸 나머지를 경로 변수에 저장.
	[self setBmspath:[path stringByDeletingLastPathComponent]];
	
	// bmsString을 줄 단위로 잘라서 배열에 저장한다. 나중 처리를 위해
	arrayOfBmsTextLines = [rawBmsText componentsSeparatedByString:@"\n"];
	
	edController = aEDController;
	// BMS 분석 코드 호출
	[self _parse];
	
	// 부모 초기화 코드 호출
	[super init];
	
	return self;
}

/**
 소멸자(에 해당하는 메서드).
 */
- (void)dealloc
{
	[notes release];
	[super dealloc];
}

/**
 @brief BMS파일을 분석하는 내부 메서드.
 
 수행순서:
 - 모든 라인을 읽을 때까지 반복
 -# 빈 줄 또는 의미없는 줄 건너뛰기
 -# #WAV 명령 처리
 -# #BMP 명령 처리
 -# 그 외의 명령 처리
 -# 노트 데이터 처리
 -# 기타 처리
 */
- (void)_parse
{
	NSScanner *scanner;
	NSString *firstTokenOfLine, *lastTokenOfLine, *args;
	unsigned int num, result;
	unsigned int numberOfNotes;
	BMSNote *arrayOfNotes;
	NSMutableArray *mutableNotes = [[NSMutableArray alloc] init];
	
	void readWAVDirective(NSString *aLineOfBmsText)
	{
		lastTokenOfLine = [aLineOfBmsText substringFromIndex:4];	// 앞에서 4글자(#WAV) 자르고 뒷부분만
		scanner = [NSScanner scannerWithString:lastTokenOfLine];	// 스캐너 생성
		[scanner scanHexInt:&num];
		// NOTE: 개행문자가 플랫폼에 따라 다를 수 있다는걸 확인해두기. \n, \r, \r\n, \n\r, ...
		[scanner scanUpToString:@"\r" intoString:&args]; // 개행문자가 나올때까지 나머지 읽기
		
		result = FMOD_System_CreateSound([edController soundSystem],
										 [[[self bmspath] stringByAppendingPathComponent:args] UTF8String],
										 FMOD_HARDWARE, 0, &audioFiles[num]);
		result = FMOD_Sound_SetMode(audioFiles[num], FMOD_LOOP_OFF);
	}
	
	void readBMPDirective(NSString *aLineOfBmsText)
	{
		lastTokenOfLine = [aLineOfBmsText substringFromIndex:4];
		scanner = [NSScanner scannerWithString:lastTokenOfLine];
		[scanner scanHexInt:&num];
		[scanner scanUpToString:@"\r" intoString:&args];
		if(num)
			animationFiles[num] = args;
	}
	
	void readNoteData(NSString *aLineOfBmsText)
	{
		/*
		 노트 데이터 처리.
		 ex: #05701:0013001300000000
		 # / 057 / 0 / 1 / : / 00 13 00 13 00 00 00 00
		 뒷부분 노트 나열은 길이가 가변적이나, 기본적으로 2의 배수x2의 길이다.
		 예를 들면 한 마디에 노트 4개가 있다면 #00000:11223344 이런 식이다.
		 */
		unsigned int noteValue;
		char *noteStringAsUTF8 = (char*)malloc(255);
		
		// sscanf를 쓰기 위해 NSString을 일반 문자열로 변환하고, 서식에 맞게 잘라 변수값을 대입한다.
		unsigned int track, type, channel;
		sscanf([aLineOfBmsText UTF8String], "#%03d%1d%1d:%s\n", &track, &type, &channel, noteStringAsUTF8);
		
		// 구한 노트 데이터 부분을 해석한다.
		NSString *noteStringAsNSString = [NSString stringWithUTF8String:noteStringAsUTF8];
		numberOfNotes = [noteStringAsNSString length] / 2;
		NSString *singleNote;
		for(int i=0; i < numberOfNotes; i++)
		{
			singleNote = [noteStringAsNSString substringWithRange:NSMakeRange(i*2, 2)];
			sscanf([singleNote UTF8String], "%02x", &noteValue);
			/* 노트값이 0이 아니고,
			 노트가 이벤트이고 1번 채널(BGM)이거나 노트가 플레이어블 객체일 경우 */
			if(noteValue &&
			   ((type == 0 && channel == 1) || type == 1))
			{
				arrayOfNotes = [[BMSNote alloc] init];
				[arrayOfNotes setAppearanceTime: (((double)i/numberOfNotes) + track) * timePerTrack];
				[arrayOfNotes setIndexOfInstrument:noteValue];
				[mutableNotes addObject:arrayOfNotes];
				[arrayOfNotes release];
			}
		}
	}
	
	NSLog(@"BMS 분석을 시작합니다.");
	// BMS 텍스트 배열의 enumerator를 생성
	NSEnumerator *bmsEnumerator = [arrayOfBmsTextLines objectEnumerator];
	
	NSString *aLineOfBmsText;
		
	@try {
		while( (aLineOfBmsText = [bmsEnumerator nextObject]) )			// 줄 단위로 처리
		{
			if(![aLineOfBmsText hasPrefix:@"#"]) continue;		// #로 시작하지 않으면 다 무시(빈 줄 포함)
			
			// 줄의 처음이 #WAV로 시작할 때의 처리(소리 파일 읽기)
			if([aLineOfBmsText hasPrefix:@"#WAV"])
				readWAVDirective(aLineOfBmsText);
			// BMP파일 읽기
			else if([aLineOfBmsText hasPrefix:@"#BMP"])
				readBMPDirective(aLineOfBmsText);
			// 노트 읽기
			else if([aLineOfBmsText characterAtIndex:6] == ':')
				readNoteData(aLineOfBmsText);
			// 나머지 잡다한 것들
			else
			{
				scanner = [NSScanner scannerWithString:aLineOfBmsText];		// 줄 데이터로 스캐너 생성
				[scanner scanUpToString:@" " intoString:&firstTokenOfLine];	// 첫 공백문자 이전 부분을 잘라낸다.
				
				// 지시어(directive)를 구분해서 알맞은 작업을 처리
				if([firstTokenOfLine isEqual:@"#STAGEFILE"])
				{
					// 뒷부분을 잘라낸 뒤 stageFile에 절대경로를 저장한다.
					[scanner scanUpToString:@"\n" intoString:&args];
					stageFile = [[self bmspath] stringByAppendingPathComponent:args];
				}
				else if([firstTokenOfLine isEqual:@"#TITLE"])
					[scanner scanUpToString:@"\n" intoString:&title];
				else if([firstTokenOfLine isEqual:@"#BANNER"])
					[scanner scanUpToString:@"\n" intoString:&banner];
				else if([firstTokenOfLine isEqual:@"#BPM"])
				{
					[scanner scanInt:&BPM];		// 다음에 오는 문자가 숫자여야 제대로 작동한다.
					timePerTrack = 240.0 / BPM;
				}
			}
		}
	}
	@catch(NSException *e) {
		NSLog(@"Exception:%@", [e description]);
	}
	
	BMSNote *temporaryNote;
	FMOD_SOUND** fmodAudios = [self audioFiles];
	unsigned int k = 0;
	
	// 각 노트의 Instrument 번호를 토대로 올바른 소리 파일을 연결한다.
	NSEnumerator *e = [mutableNotes objectEnumerator];
	while( (temporaryNote = [e nextObject]) )
	{
		k = [temporaryNote indexOfInstrument];
		FMOD_SOUND *fs = fmodAudios[k];
		[temporaryNote setAudio:fs];
	}
	
	notes = [mutableNotes copy];
	
	[mutableNotes release];
}

/**
 fromTime과 toTime 사이의 시간대에 존재하는 노트들을 목록화한다.
 @param fromTime 시작 시간
 @param toTime 끝 시간
 */
- (NSArray *)notesWithTimeFrom:(CFAbsoluteTime)fromTime to:(CFAbsoluteTime)toTime
{
	BMSNote *aNote;
	NSMutableArray *mutableArray = [[NSMutableArray alloc] init];
	NSEnumerator *e = [notes objectEnumerator];
	while( (aNote = [e nextObject]) )
	{
		if([aNote appearanceTime] > fromTime && [aNote appearanceTime] <= toTime)
			[mutableArray addObject:aNote];
	}
	return mutableArray;
}

- (FMOD_SOUND **)audioFiles
{
	return audioFiles;
}

@end
